import argparse
import glob
import os
import sqlite3
from collections import OrderedDict
from http.server import BaseHTTPRequestHandler, HTTPServer
from subprocess import check_output
from uuid import uuid4

import yaml
from jinja2 import Template
from xfel_calibrate.settings import (free_nodes_cmd, preempt_nodes_cmd,
                                     reservation)


class LimitedSizeDict(OrderedDict):
    def __init__(self, *args, **kwds):
        self.size_limit = kwds.pop("size_limit", None)
        OrderedDict.__init__(self, *args, **kwds)
        self._check_size_limit()

    def __setitem__(self, key, value):
        OrderedDict.__setitem__(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.size_limit is not None:
            while len(self) > self.size_limit:
                self.popitem(last=False)


config = None
pdf_queue = LimitedSizeDict(size_limit=50)


# HTTPRequestHandler class
class RequestHandler(BaseHTTPRequestHandler):
    conf_was_init = False

    def init_config(self):

        global config

        self.nodes_avail_res_cmd = config["shell-commands"]["nodes-avail-res"]
        self.total_jobs_cmd = config["shell-commands"]["total-jobs"]
        self.upex_jobs_cmd = config["shell-commands"]["upex-jobs"]
        self.upex_prefix = config["shell-commands"]["upex-prefix"]
        self.tail_log_cmd = config["shell-commands"]["tail-log"]
        self.cat_log_cmd = config["shell-commands"]["cat-log"]
        self.mappings = config["mappings"]
        self.run_candidates = config["run-candidates"]

        self.templates = {}
        for template, tfile in config["templates"].items():
            with open(tfile, "r") as tf:
                self.templates[template] = tf.read()
        global pdf_queue
        self.pdf_queue = pdf_queue
        self.conf_was_init = True

    def do_GET(self):

        if not self.conf_was_init:
            self.init_config()
        # Send response status code
        self.send_response(200)
        if self.path == "/serve_overview.css":
            self.send_header('Content-type', 'text/css')
            self.end_headers()
            for s in self.templates["css"].split("\n"):
                self.wfile.write(s.encode())
            return
        if "pdf?" in self.path:
            puuid = self.path.split("?")[1]
            fpath = self.pdf_queue.get(puuid, None)
            if fpath is None:
                return
            self.send_header('Content-type', 'application/pdf')
            self.end_headers()
            with open(fpath, "rb") as f:
                self.wfile.write(f.read())
            return

        # Send headers
        self.send_header('Content-type', 'text/html')
        self.end_headers()

        # Send message back to client
        # general maxwell stats
        free = int(check_output(free_nodes_cmd, shell=True).decode('utf8'))
        preempt = int(check_output(
            preempt_nodes_cmd, shell=True).decode('utf8'))
        nodes_avail_general = free + preempt
        ret = check_output(self.nodes_avail_res_cmd.format(reservation),
                           shell=True).decode('utf8')
        ures = ret.split("/")
        if len(ures) == 2:
            used, reserved = ures
        else:
            used = 0
            reserved = 0
        nodes_avail_gr = "{}/{}".format(int(reserved) - int(used),
                                        reserved)
        total_jobs_running = check_output(self.total_jobs_cmd, shell=True)
        total_jobs_running = total_jobs_running.decode('utf8').split("/")[0]

        upex_res = [r for r in
                    check_output(self.upex_jobs_cmd,
                                 shell=True).decode('utf8').split()
                    if self.upex_prefix in r]

        upex_reservations = {}
        for res in upex_res:
            nodes_avail_res = check_output(
                self.nodes_avail_res_cmd.format(res),
                shell=True).decode('utf8')
            used, reserved = nodes_avail_res.split("/")
            nodes_avail_res = "{}/{}".format(int(reserved) - int(used),
                                             reserved)
            upex_reservations[res] = nodes_avail_res

        recommendation = "DON'T SUBMIT TO RESERVATION"
        if nodes_avail_general < int(reserved) - int(used):
            "CONSIDER SUBMITTING TO RESERVATION"

        tmpl = Template(self.templates["maxwell-status"])
        maxwell_status_r = tmpl.render(nodes_avail_general=nodes_avail_general,
                                       nodes_avail_general_res=nodes_avail_gr,
                                       total_jobs_running=total_jobs_running,
                                       upex_reservations=upex_reservations,
                                       recommendation=recommendation)

        last_n_lines = check_output(self.tail_log_cmd,
                                    shell=True).decode('utf8').split("\n")
        last_n_lines = [l for l in last_n_lines
                        if "Response error from MDC" not in l]
        tmpl = Template(self.templates["log-output"])
        log_output_r = tmpl.render(logout="<br>".join(last_n_lines[::-1]))

        last_n_lines = check_output(self.cat_log_cmd,
                                    shell=True).decode('utf8').split("\n")[
                       ::-1]

        last_chars = {}
        for l in last_n_lines:
            if "DARK" in l:
                ls = l.split()
                if "DARK" not in ls:
                    continue
                detector = ls[ls.index("DARK") - 1]

                if "--instrument" in ls:
                    instrument = ls[ls.index("--instrument") + 1]
                else:
                    if "--db-module" in ls:
                        instrument = ls[ls.index("--db-module") + 1]
                    elif detector == "PNCCD":
                        instrument = "SQS"
                    elif detector == "FASTCCD":
                        instrument = "SCS"
                    else:
                        instrument = ""

                in_folder = ls[ls.index("--in-folder") + 1]
                out_folder = ls[ls.index("--out-folder") + 1]
                runs = []
                for rc in self.run_candidates:
                    if rc in ls:
                        runs.append(ls[ls.index(rc) + 1])

                if f"{instrument}-{detector}" in last_chars:
                    continue

                requested = "{} {}".format(ls[0], ls[1])

                pdfs = glob.glob(f"{out_folder}/*.pdf")
                pdfs = {p.split("/")[-1]: p for p in pdfs}
                fpdfs = []
                if "xfel.pdf" in pdfs:
                    del pdfs["xfel.pdf"]
                if len(pdfs):
                    # pdfs = ", ".join(pdfs)
                    for pdf, p in pdfs.items():
                        puuid = uuid4().hex
                        self.pdf_queue[puuid] = p
                        host = config["server-config"]["host"]
                        port = config["server-config"]["port"]
                        fpdfs.append(
                            (pdf, f"http://{host}:{port}/pdf?{puuid}"))
                pdfs = fpdfs
                tsize = 0
                for run in runs:
                    run = int(run)
                    if detector not in self.mappings:
                        continue
                    for mp in self.mappings[detector]:
                        for f in glob.glob(
                                f"{in_folder}/r{run:04d}/*{mp}*.h5"):
                            tsize += os.stat(f).st_size
                last_injected = ""
                constant_valid_from = ""
                key = f"{instrument}-{detector}"
                last_chars[key] = {"in_path": in_folder,
                                   "out_path": out_folder,
                                   "runs": runs,
                                   "pdfs": pdfs,
                                   "size": "{:0.1f} GB".format(tsize / 1e9),
                                   "requested": requested,
                                   "device_type": detector,
                                   "last_valid_from": constant_valid_from}

        tmpl = self.templates["last-characterizations"]
        last_characterizations_r = Template(tmpl).render(char_runs=last_chars)

        conn = sqlite3.connect(config['web-service']['job-db']).cursor()
        conn.execute("SELECT * FROM jobs")
        running_jobs = {}
        for r in conn.fetchall():
            rid, jobid, proposal, run, flg, status = r
            run = int(run)
            rjobs = running_jobs.get("r{:04d}".format(run),
                                     {'proposal': proposal, 'statii': []})
            rjobs["statii"].append((flg, status))
            running_jobs["r{:04d}".format(run)] = rjobs

        tmpl = self.templates["running-jobs"]
        running_jobs_r = Template(tmpl).render(running_jobs=running_jobs)

        tmpl = Template(self.templates["main-doc"])
        message = tmpl.render(maxwell_status=maxwell_status_r,
                              log_output=log_output_r,
                              last_characterizations=last_characterizations_r,
                              running_jobs=running_jobs_r)
        # Write content as utf-8 data
        self.wfile.write(bytes(message, "utf8"))
        return


def run(configfile, port=8008):
    print('reading config file')
    with open(configfile, "r") as cf:
        global config
        config = yaml.load(cf.read())
    print('starting server...')
    sconfig = config["server-config"]
    server_address = (sconfig["host"], sconfig["port"])
    httpd = HTTPServer(server_address, RequestHandler)
    print('running server...')
    httpd.serve_forever()


parser = argparse.ArgumentParser(
    description='Start the overview server')
parser.add_argument('--config', type=str, default="serve_overview.yaml")
if __name__ == "__main__":
    args = vars(parser.parse_args())
    run(args["config"])
